@@ -28,7 +28,7 @@
  * @short_description: Base class for actors which contain multiple child
  * actors.
  *
- * #ClutterGroup is an Actor which can contain multiple child actors.
+ * A #ClutterGroup is an Actor which can contain multiple child actors.
  */
 
 #include "config.h"
@@ -67,19 +67,6 @@ clutter_group_paint (ClutterActor *actor)
 
   glPushMatrix();
 
-  /* Translate if parent ( i.e not stage window ).
-  */
-  if (clutter_actor_get_parent (actor) != NULL)
-    {
-      ClutterGeometry geom;      
-
-      clutter_actor_get_geometry (actor, &geom);
-
-      if (geom.x != 0 && geom.y != 0)
-	glTranslatef(geom.x, geom.y, 0.0);
-
-    }
-
   for (child_item = self->priv->children;
        child_item != NULL;
        child_item = child_item->next)
@@ -99,60 +86,15 @@ static void
 clutter_group_request_coords (ClutterActor    *self,
 			      ClutterActorBox *box)
 {
-  ClutterGroup   *group = CLUTTER_GROUP(self);
-  guint           cwidth, cheight, width ,height;
   ClutterActorBox cbox;
 
   clutter_actor_allocate_coords (self, &cbox);
 
-  cwidth  = cbox.x2 - cbox.x1;
-  cheight = cbox.y2 - cbox.y1;
-
-  /* g_print("cbox x2: %i x1 %i\n", cbox.x2, cbox.x1); */
-
-  width  = box->x2 - box->x1;
-  height = box->y2 - box->y1;
-
-  /* FIXME: below needs work */
-  if (cwidth != width || cheight != height)
-    {
-      GList          *child_item;
-
-      for (child_item = group->priv->children;
-	   child_item != NULL;
-	   child_item = child_item->next)
-	{
-	  ClutterActor   *child = child_item->data;
-	  ClutterActorBox tbox;
-	  gint            nx, ny;
-	  gint            twidth, theight, nwidth, nheight; 
-
-	  g_assert (child != NULL);
-
-	  clutter_actor_allocate_coords (child, &tbox);
-
-	  twidth  = tbox.x2 - tbox.x1;
-	  theight = tbox.y2 - tbox.y1;
-	  
-	  /* g_print("getting ps %ix%i\n", tbox.x1, tbox.y1); */
-
-	  nwidth = ( width * twidth ) / cwidth;
-	  nheight = ( height * theight ) / cheight;
-
-	  nx = ( nwidth  * tbox.x1 ) / twidth ;
-
-	  /* g_print("n: %i t %i  x1: %i\n", nwidth, twidth, tbox.x1); */
-
-	  ny = ( nheight  * tbox.y1 ) / theight;
-
-	  /* g_print("n: %i t %i  x1: %i\n", nheight, theight, tbox.y1); */
-
-	  clutter_actor_set_position (child, nx, ny);
-	  clutter_actor_set_size (child, nwidth, height);
-
-	  /* g_print("size to +%i+%x %ix%i\n", nx, ny, nwidth, nheight); */
-	}
-    }
+  /* Only positioning works.
+   * Sizing requests fail, use scale() instead 
+  */
+  box->x2 = box->x1 + (cbox.x2 - cbox.x1);
+  box->y2 = box->y1 + (cbox.y2 - cbox.y1);
 }
 
 static void
@@ -166,30 +108,30 @@ clutter_group_allocate_coords (ClutterActor    *self,
 
   child_item = priv->children;
 
+  /* FIXME: Cache these values */
+
   if (child_item)
     {
       do 
 	{
 	  ClutterActor *child = CLUTTER_ACTOR(child_item->data);
-	      
+
+	  /* Once added we include in sizing - doesn't matter if visible */
 	  /* if (CLUTTER_ACTOR_IS_VISIBLE (child)) */
 	    {
 	      ClutterActorBox cbox;
 
 	      clutter_actor_allocate_coords (child, &cbox);
 	      
-	      /*
-	      if (box->x1 == 0 || cbox.x1 < box->x1)
-		box->x1 = cbox.x1;
-
-	      if (box->y1 == 0 || cbox.y1 < box->y1)
-		box->y1 = cbox.y1;
-	      */		
-
+	      /* Ignore any children with offscreen ( negaive )
+               * positions.
+	       *
+               * Also x1 and x2 will be set by parent caller.
+	      */
 	      if (box->x2 == 0 || cbox.x2 > box->x2)
 		box->x2 = cbox.x2;
 
-	      if (box->y2 == 0 || cbox.y2 < box->y2)
+	      if (box->y2 == 0 || cbox.y2 > box->y2)
 		box->y2 = cbox.y2;
 	    }
 	}
@@ -200,16 +142,12 @@ clutter_group_allocate_coords (ClutterActor    *self,
 static void 
 clutter_group_dispose (GObject *object)
 {
-  ClutterGroup *self = CLUTTER_GROUP(object); 
-
-  if (self->priv)
-    {
-      /* FIXME: Do we need to actually free anything here ?
-       *        Children ref us so this wont get called till
-       *        they are all removed.
-      */
-    }
+  ClutterGroup *self = CLUTTER_GROUP (object);
 
+  clutter_group_foreach (self,
+		  	 CLUTTER_CALLBACK (clutter_actor_destroy),
+			 NULL);
+  
   G_OBJECT_CLASS (clutter_group_parent_class)->dispose (object);
 }
 
@@ -217,37 +155,42 @@ clutter_group_dispose (GObject *object)
 static void 
 clutter_group_finalize (GObject *object)
 {
-  ClutterGroup *group = CLUTTER_GROUP (object);
-
-  /* XXX - if something survives ::dispose then there's something
-   * wrong; but, at least, we won't leak stuff around.
-   */
-  if (group->priv->children)
-    {
-      g_list_foreach (group->priv->children, (GFunc) g_object_unref, NULL);
-      g_list_free (group->priv->children);
-    }
-  
   G_OBJECT_CLASS (clutter_group_parent_class)->finalize (object);
 }
 
+static void
+clutter_group_real_show_all (ClutterActor *actor)
+{
+  clutter_group_foreach (CLUTTER_GROUP (actor),
+                         CLUTTER_CALLBACK (clutter_actor_show_all),
+                         NULL);
+  clutter_actor_show (actor);
+}
+
+static void
+clutter_group_real_hide_all (ClutterActor *actor)
+{
+  clutter_actor_hide (actor);
+  clutter_group_foreach (CLUTTER_GROUP (actor),
+                         CLUTTER_CALLBACK (clutter_actor_hide_all),
+                         NULL);
+}
+
 static void
 clutter_group_class_init (ClutterGroupClass *klass)
 {
-  GObjectClass        *object_class = G_OBJECT_CLASS (klass);
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
   ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
 
-  actor_class->paint      = clutter_group_paint;
-  /*
-  actor_class->show       = clutter_group_show_all;
-  actor_class->hide       = clutter_group_hide_all;
-  */
+  actor_class->paint           = clutter_group_paint;
+  actor_class->show_all        = clutter_group_real_show_all;
+  actor_class->hide_all        = clutter_group_real_hide_all;
   actor_class->request_coords  = clutter_group_request_coords;
   actor_class->allocate_coords = clutter_group_allocate_coords;
 
   /* GObject */
-  object_class->finalize     = clutter_group_finalize;
-  object_class->dispose      = clutter_group_dispose;
+  object_class->finalize = clutter_group_finalize;
+  object_class->dispose = clutter_group_dispose;
 
   group_signals[ADD] =
     g_signal_new ("add",
@@ -285,7 +228,7 @@ clutter_group_init (ClutterGroup *self)
  *
  * returns a new #ClutterGroup
  **/
-ClutterGroup*
+ClutterActor *
 clutter_group_new (void)
 {
   return g_object_new (CLUTTER_TYPE_GROUP, NULL);
@@ -297,17 +240,57 @@ clutter_group_new (void)
  * 
  * Get a list containing all actors contained in the group.
  * 
- * Return value: A GList containing child #ClutterActors.
- **/
+ * Return value: A GList containing child #ClutterActors.  You
+ *   should free the returned list using g_list_free() when
+ *   finished using it.
+ */
 GList*
 clutter_group_get_children (ClutterGroup *self)
 {
-  /* FIXME: remane get_actors() */
   g_return_val_if_fail (CLUTTER_IS_GROUP (self), NULL);
 
   return g_list_copy(self->priv->children);
 }
 
+/**
+ * clutter_group_get_n_children:
+ * @self: A #ClutterGroup
+ *
+ * Gets the number of actors held in the group.
+ * 
+ * Return value: The number of child actors held in the group.
+ * 
+ * Since: 0.2
+ **/
+gint
+clutter_group_get_n_children (ClutterGroup *self)
+{
+  g_return_val_if_fail (CLUTTER_IS_GROUP (self), 0);
+
+  return g_list_length (self->priv->children);
+}
+
+/**
+ * clutter_group_get_nth_child:
+ * @self: A #ClutterGroup
+ * @index: the position of the requested actor. 
+ * 
+ * Gets a groups child held at position index in stack.
+ *
+ * Return value: A Clutter actor or NULL if index is invalid.
+ *
+ * Since: 0.2
+ **/
+ClutterActor *
+clutter_group_get_nth_child (ClutterGroup *self,
+			     gint          index)
+{
+  g_return_val_if_fail (CLUTTER_IS_GROUP (self), NULL);
+
+  return g_list_nth_data (self->priv->children, index);
+}
+
+
 /**
  * clutter_group_foreach:
  * @self: A #ClutterGroup
@@ -343,36 +326,32 @@ clutter_group_foreach (ClutterGroup      *self,
  * clutter_group_show_all:
  * @self: A #ClutterGroup
  * 
- * Show all child actors of the group. Note, does not recurse.
- **/
+ * Show all child actors of the group, like clutter_actor_show_all().
+ *
+ * @Deprecated: Use clutter_actor_show_all() instead.
+ */
 void
 clutter_group_show_all (ClutterGroup *self)
 {
   g_return_if_fail (CLUTTER_IS_GROUP (self));
 
-  clutter_actor_show(CLUTTER_ACTOR(self));
-
-  g_list_foreach (self->priv->children,
-		  (GFunc)clutter_actor_show,
-		  NULL);
+  clutter_actor_show_all (CLUTTER_ACTOR (self));
 }
 
 /**
  * clutter_group_hide_all:
  * @self: A #ClutterGroup
  * 
- * Hide all child actors of the group. Note, does not recurse.
- **/
+ * Hide all child actors of the group, like clutter_actor_hide_all().
+ *
+ * @Deprecated: Use clutter_actor_hide_all() instead
+ */
 void
 clutter_group_hide_all (ClutterGroup *self)
 {
   g_return_if_fail (CLUTTER_IS_GROUP (self));
 
-  clutter_actor_hide(CLUTTER_ACTOR(self));
-
-  g_list_foreach (self->priv->children,
-		  (GFunc)clutter_actor_hide,
-		  NULL);
+  clutter_actor_hide_all (CLUTTER_ACTOR (self));
 }
 
 /**
@@ -383,10 +362,12 @@ clutter_group_hide_all (ClutterGroup *self)
  * Adds a new child #ClutterActor to the #ClutterGroup.
  **/
 void
-clutter_group_add (ClutterGroup   *self,
+clutter_group_add (ClutterGroup *self,
 		   ClutterActor *actor)
 {
   ClutterActor *parent;
+
+  /* FIXME: add() needs to be somehow overidden */
   
   g_return_if_fail (CLUTTER_IS_GROUP (self));
   g_return_if_fail (CLUTTER_IS_ACTOR (actor));
@@ -403,19 +384,21 @@ clutter_group_add (ClutterGroup   *self,
       return;
     }
 
-  self->priv->children = g_list_append (self->priv->children, actor);
-  /* below refs */
-  clutter_actor_set_parent (actor, CLUTTER_ACTOR(self));
   g_object_ref (actor);
 
+  self->priv->children = g_list_append (self->priv->children, actor);
+  clutter_actor_set_parent (actor, CLUTTER_ACTOR (self));
+  
   clutter_group_sort_depth_order (self); 
 
   g_signal_emit (self, group_signals[ADD], 0, actor);
+
+  g_object_unref (actor);
 }
 
 /**
  * clutter_group_add_many_valist:
- * @group: a #ClutterGroup
+ * @self: a #ClutterGroup
  * @first_actor: the #ClutterActor actor to add to the group
  * @args: the actors to be added
  *
@@ -423,19 +406,19 @@ clutter_group_add (ClutterGroup   *self,
  * function inside bindings.
  */
 void
-clutter_group_add_many_valist (ClutterGroup   *group,
+clutter_group_add_many_valist (ClutterGroup *self,
 			       ClutterActor *first_actor,
-			       va_list         args)
+			       va_list       args)
 {
   ClutterActor *actor;
   
-  g_return_if_fail (CLUTTER_IS_GROUP (group));
+  g_return_if_fail (CLUTTER_IS_GROUP (self));
   g_return_if_fail (CLUTTER_IS_ACTOR (first_actor));
 
   actor = first_actor;
   while (actor)
     {
-      clutter_group_add (group, actor);
+      clutter_group_add (self, actor);
       actor = va_arg (args, ClutterActor *);
     }
 }
@@ -450,7 +433,7 @@ clutter_group_add_many_valist (ClutterGroup   *group,
  * equivalent to calling clutter_group_add() for each member of the list.
  */
 void
-clutter_group_add_many (ClutterGroup   *self,
+clutter_group_add_many (ClutterGroup *self,
 		        ClutterActor *first_actor,
 			...)
 {
@@ -487,11 +470,17 @@ clutter_group_remove (ClutterGroup *self,
 		 g_type_name (G_OBJECT_TYPE (self)));
       return;
     }
+  
+  g_object_ref (actor);
 
   self->priv->children = g_list_remove (self->priv->children, actor);
-  clutter_actor_set_parent (actor, NULL);
+  clutter_actor_unparent (actor);
   
   g_signal_emit (self, group_signals[REMOVE], 0, actor);
+
+  if (CLUTTER_ACTOR_IS_VISIBLE (CLUTTER_ACTOR (self)))
+    clutter_actor_queue_redraw (CLUTTER_ACTOR (self));
+
   g_object_unref (actor);
 }
 
