@@ -26,8 +26,6 @@
 #ifndef _HAVE_CLUTTER_PRIVATE_H
 #define _HAVE_CLUTTER_PRIVATE_H
 
-#include <config.h>
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -45,6 +43,8 @@
 
 #include <pango/pangoft2.h>
 
+G_BEGIN_DECLS
+
 typedef struct _ClutterMainContext ClutterMainContext;
 
 struct _ClutterMainContext
@@ -70,4 +70,27 @@ struct _ClutterMainContext
 #define CLUTTER_CONTEXT()	(clutter_context_get_default ())
 ClutterMainContext *clutter_context_get_default (void);
 
+const gchar *clutter_vblank_method (void);
+
+typedef enum {
+  CLUTTER_ACTOR_UNUSED_FLAG = 0,
+
+  CLUTTER_ACTOR_IN_DESTRUCTION = 1 << 0,
+  CLUTTER_ACTOR_IS_TOPLEVEL    = 1 << 1,
+  CLUTTER_ACTOR_IN_REPARENT    = 1 << 2
+} ClutterPrivateFlags;
+
+#define CLUTTER_PRIVATE_FLAGS(a)	 (CLUTTER_ACTOR ((a))->private_flags)
+#define CLUTTER_SET_PRIVATE_FLAGS(a,f)	 G_STMT_START{ (CLUTTER_PRIVATE_FLAGS (a) |= (f)); }G_STMT_END
+#define CLUTTER_UNSET_PRIVATE_FLAGS(a,f) G_STMT_START{ (CLUTTER_PRIVATE_FLAGS (a) &= ~(f)); }G_STMT_END
+
+#define CLUTTER_PARAM_READABLE  \
+        G_PARAM_READABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB
+#define CLUTTER_PARAM_WRITABLE  \
+        G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB
+#define CLUTTER_PARAM_READWRITE \
+        G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |G_PARAM_STATIC_BLURB
+
+G_END_DECLS
+
 #endif
