@@ -39,7 +39,8 @@
 #include "clutter-util.h"
 #include "clutter-marshal.h"
 #include "clutter-enum-types.h"
-#include "clutter-private.h" 	/* for DBG */
+#include "clutter-private.h"
+#include "clutter-debug.h"
 
 #include <GL/glx.h>
 #include <GL/gl.h>
@@ -70,6 +71,7 @@ struct _ClutterStagePrivate
   guint         want_fullscreen : 1;
   guint         want_offscreen  : 1;
   guint         hide_cursor     : 1;
+  guint         is_foreign_xwin : 1;
 };
 
 enum
@@ -236,17 +238,29 @@ sync_gl_viewport (ClutterStage *stage)
 static void
 clutter_stage_show (ClutterActor *self)
 {
-  if (clutter_stage_get_xwindow (CLUTTER_STAGE(self)))
-    XMapWindow (clutter_xdisplay(), 
-		clutter_stage_get_xwindow (CLUTTER_STAGE(self)));
+  ClutterActorClass *parent_class;
+
+  parent_class = CLUTTER_ACTOR_CLASS (clutter_stage_parent_class);
+  if (parent_class->show)
+    parent_class->show (self);
+
+  if (clutter_stage_get_xwindow (CLUTTER_STAGE (self)))
+    XMapWindow (clutter_xdisplay (), 
+		clutter_stage_get_xwindow (CLUTTER_STAGE (self)));
 }
 
 static void
 clutter_stage_hide (ClutterActor *self)
 {
-  if (clutter_stage_get_xwindow (CLUTTER_STAGE(self)))
-    XUnmapWindow (clutter_xdisplay(), 
-		  clutter_stage_get_xwindow (CLUTTER_STAGE(self)));
+  ClutterActorClass *parent_class;
+
+  parent_class = CLUTTER_ACTOR_CLASS (clutter_stage_parent_class);
+  if (parent_class->hide)
+    parent_class->hide (self);
+
+  if (clutter_stage_get_xwindow (CLUTTER_STAGE (self)))
+    XUnmapWindow (clutter_xdisplay (), 
+		  clutter_stage_get_xwindow (CLUTTER_STAGE (self)));
 }
 
 static void
@@ -262,8 +276,6 @@ clutter_stage_unrealize (ClutterActor *actor)
 
   if (priv->want_offscreen)
     {
-      
-
       if (priv->glxpixmap)
 	{
 	  glXDestroyGLXPixmap (clutter_xdisplay(), priv->glxpixmap);
@@ -278,16 +290,21 @@ clutter_stage_unrealize (ClutterActor *actor)
     }
   else
     {
-      if (clutter_stage_get_xwindow (CLUTTER_STAGE(actor)))
+      if (!priv->is_foreign_xwin && priv->xwin != None)
 	{
 	  XDestroyWindow (clutter_xdisplay(), priv->xwin);
 	  priv->xwin = None;
 	}
+      else
+        priv->xwin = None;
     }
 
   glXMakeCurrent(clutter_xdisplay(), None, NULL);
-  glXDestroyContext (clutter_xdisplay(), priv->gl_context);
-  priv->gl_context = None;
+  if (priv->gl_context != None)
+    {
+      glXDestroyContext (clutter_xdisplay(), priv->gl_context);
+      priv->gl_context = None;
+    }
 }
 
 static void
@@ -325,7 +342,7 @@ clutter_stage_realize (ClutterActor *actor)
 	  return;
 	}
 
-      if (priv->gl_context)
+      if (priv->gl_context != None)
 	glXDestroyContext (clutter_xdisplay(), priv->gl_context);
 
       priv->xpixmap = XCreatePixmap (clutter_xdisplay(),
@@ -383,15 +400,37 @@ clutter_stage_realize (ClutterActor *actor)
 	  GLX_RED_SIZE, 1,
 	  GLX_GREEN_SIZE, 1,
 	  GLX_BLUE_SIZE, 1,
+	  GLX_STENCIL_SIZE, 1,
 	  0
 	};
 
       if (priv->xvisinfo)
-	XFree(priv->xvisinfo);
+	{
+	  XFree(priv->xvisinfo);
+	  priv->xvisinfo = None;
+	}
+#if 0
+      /* Attempted fix at GTK 'white' textures - made no difference :( */
+      if (priv->is_foreign_xwin && priv->xwin != None)
+	{
+	  XWindowAttributes win_attr;
+	  XVisualInfo       vis_info;
+	  int               n;
 
-      priv->xvisinfo = glXChooseVisual (clutter_xdisplay(),
-					clutter_xscreen(),
-					gl_attributes);
+	  XGetWindowAttributes (clutter_xdisplay(), priv->xwin, &win_attr);
+	  vis_info.screen   = clutter_xscreen();
+	  vis_info.visualid = XVisualIDFromVisual (win_attr.visual);
+	  priv->xvisinfo = XGetVisualInfo (clutter_xdisplay(),
+					   VisualScreenMask|VisualIDMask, 
+					   &vis_info, &n);
+
+	  printf("made %li\n", priv->xvisinfo);
+	}
+#endif
+      if (priv->xvisinfo == None)
+	priv->xvisinfo = glXChooseVisual (clutter_xdisplay(),
+					  clutter_xscreen(),
+					  gl_attributes);
       if (!priv->xvisinfo)
 	{
 	  g_critical ("Unable to find suitable GL visual.");
@@ -399,14 +438,42 @@ clutter_stage_realize (ClutterActor *actor)
 	  return;
 	}
 
+      CLUTTER_NOTE(GL, "visual id's %li vs %li",
+		   priv->xvisinfo->visualid,
+		   XVisualIDFromVisual(DefaultVisual(clutter_xdisplay(),
+						     clutter_xscreen())));
+
+
       if (priv->xwin == None)
-	priv->xwin = XCreateSimpleWindow(clutter_xdisplay(),
-					 clutter_root_xwindow(),
-					 0, 0,
-					 priv->xwin_width, priv->xwin_height,
-					 0, 0, 
-					 WhitePixel(clutter_xdisplay(), 
-						    clutter_xscreen()));
+	{
+	  XSetWindowAttributes swa;
+
+	  if (priv->xvisinfo->visualid 
+	      == XVisualIDFromVisual(DefaultVisual(clutter_xdisplay(),
+						       clutter_xscreen())))
+	    {
+	      swa.colormap = DefaultColormap(clutter_xdisplay(),
+					     clutter_xscreen());
+	    }
+	  else
+	    {
+	      swa.colormap = XCreateColormap(clutter_xdisplay(),
+					     clutter_root_xwindow(), 
+					     priv->xvisinfo->visual, 
+					     AllocNone);
+	    }
+
+	  priv->xwin = XCreateWindow(clutter_xdisplay(),
+				     clutter_root_xwindow(), 
+				     0, 0, 
+				     priv->xwin_width, priv->xwin_height,
+				     0, 
+				     priv->xvisinfo->depth, 
+				     InputOutput, 
+				     priv->xvisinfo->visual,
+				     CWColormap, &swa);
+	}
+
       XSelectInput(clutter_xdisplay(), 
 		   priv->xwin, 
 		   StructureNotifyMask
@@ -422,23 +489,38 @@ clutter_stage_realize (ClutterActor *actor)
       sync_fullscreen (stage);  
       sync_cursor_visible (stage);  
 
-      if (priv->gl_context)
+      if (priv->gl_context != None)
 	glXDestroyContext (clutter_xdisplay(), priv->gl_context);
 
       priv->gl_context = glXCreateContext (clutter_xdisplay(), 
 					   priv->xvisinfo, 
 					   0, 
 					   True);
-      
+      if (priv->gl_context == None)
+	{
+	  g_critical ("Unable to create suitable GL context.");
+	  CLUTTER_ACTOR_UNSET_FLAGS (actor, CLUTTER_ACTOR_REALIZED);
+	  return;
+	}
+	
       glXMakeCurrent(clutter_xdisplay(), priv->xwin, priv->gl_context);
     }
 
-  CLUTTER_DBG("===========================================")
-  CLUTTER_DBG("GL_VENDOR: %s\n", glGetString(GL_VENDOR));
-  CLUTTER_DBG("GL_RENDERER: %s\n", glGetString(GL_RENDERER));
-  CLUTTER_DBG("GL_VERSION: %s\n", glGetString(GL_VERSION));
-  CLUTTER_DBG("GL_EXTENSIONS: %s\n", glGetString(GL_EXTENSIONS));
-  CLUTTER_DBG("===========================================")
+  CLUTTER_NOTE (GL,
+                "\n"
+		"==========================================="
+		"GL_VENDOR: %s\n"
+		"GL_RENDERER: %s\n"
+		"GL_VERSION: %s\n"
+		"GL_EXTENSIONS: %s\n"
+		"Is direct: %s\n"
+		"===========================================",
+		glGetString (GL_VENDOR),
+		glGetString (GL_RENDERER),
+		glGetString (GL_VERSION),
+		glGetString (GL_EXTENSIONS),
+		glXIsDirect(clutter_xdisplay(), priv->gl_context) ? "yes" : "no"
+		);
 
   sync_gl_viewport (stage);
 }
@@ -481,7 +563,7 @@ clutter_stage_request_coords (ClutterActor    *self,
       priv->xwin_width  = new_width;
       priv->xwin_height = new_height;
 
-      if (priv->xwin)
+      if (priv->xwin != None)
 	XResizeWindow (clutter_xdisplay(), 
 		       priv->xwin, 
 		       priv->xwin_width, 
@@ -497,11 +579,12 @@ clutter_stage_request_coords (ClutterActor    *self,
       sync_gl_viewport (stage);
     }
 
-  if (priv->xwin) /* Do we want to bother ? */
+  if (priv->xwin != None) /* Do we want to bother ? */
     XMoveWindow (clutter_xdisplay(), 
 		 priv->xwin,
 		 box->x1,
 		 box->y1);
+
 }
 
 static void 
@@ -509,7 +592,7 @@ clutter_stage_dispose (GObject *object)
 {
   ClutterStage *self = CLUTTER_STAGE (object);
 
-  if (self->priv->xwin)
+  if (self->priv->xwin != None)
     clutter_stage_unrealize (CLUTTER_ACTOR (self));
 
   G_OBJECT_CLASS (clutter_stage_parent_class)->dispose (object);
@@ -628,7 +711,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
   gobject_class->finalize     = clutter_stage_finalize;
   gobject_class->set_property = clutter_stage_set_property;
   gobject_class->get_property = clutter_stage_get_property;
-
+  
   /**
    * ClutterStage:fullscreen
    *
@@ -640,7 +723,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
 			   "Fullscreen",
 			   "Make Clutter stage fullscreen",
 			   FALSE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_OFFSCREEN,
@@ -648,7 +731,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
 			   "Offscreen",
 			   "Make Clutter stage offscreen",
 			   FALSE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
 
   g_object_class_install_property
@@ -657,7 +740,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
 			   "Hide Cursor",
 			   "Make Clutter stage cursor-less",
 			   FALSE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_COLOR,
@@ -665,8 +748,17 @@ clutter_stage_class_init (ClutterStageClass *klass)
 			 "Color",
 			 "The color of the stage",
 			 CLUTTER_TYPE_COLOR,
-			 G_PARAM_READWRITE));
+			 CLUTTER_PARAM_READWRITE));
 
+  /**
+   * ClutterStage::input-event:
+   * @stage: the actor which received the event
+   * @event: the event received
+   *
+   * The ::input-event is a signal emitted when any input event is
+   * received.  Valid input events are mouse button press and release
+   * events, and key press and release events.
+   */
   stage_signals[INPUT_EVENT] =
     g_signal_new ("input-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
@@ -676,7 +768,14 @@ clutter_stage_class_init (ClutterStageClass *klass)
 		  clutter_marshal_VOID__BOXED,
 		  G_TYPE_NONE, 1,
 		  CLUTTER_TYPE_EVENT);
-
+  /**
+   * ClutterStage::button-press-event:
+   * @stage: the actor which received the event
+   * @event: a #ClutterButtonEvent
+   *
+   * The ::button-press-event is emitted each time a mouse button
+   * is pressed on @stage.
+   */
   stage_signals[BUTTON_PRESS_EVENT] =
     g_signal_new ("button-press-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
@@ -686,6 +785,14 @@ clutter_stage_class_init (ClutterStageClass *klass)
 		  clutter_marshal_VOID__BOXED,
 		  G_TYPE_NONE, 1,
 		  CLUTTER_TYPE_EVENT);
+  /**
+   * ClutterStage::button-release-event:
+   * @stage: the actor which received the event
+   * @event: a #ClutterButtonEvent
+   *
+   * The ::button-release-event is emitted each time a mouse button
+   * is released on @stage.
+   */
   stage_signals[BUTTON_RELEASE_EVENT] =
     g_signal_new ("button-release-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
@@ -695,7 +802,15 @@ clutter_stage_class_init (ClutterStageClass *klass)
 		  clutter_marshal_VOID__BOXED,
 		  G_TYPE_NONE, 1,
 		  CLUTTER_TYPE_EVENT);
-  stage_signals[BUTTON_PRESS_EVENT] =
+  /**
+   * ClutterStage::key-press-event:
+   * @stage: the actor which received the event
+   * @event: a #ClutterKeyEvent
+   *
+   * The ::key-press-event is emitted each time a keyboard button
+   * is pressed on @stage.
+   */
+  stage_signals[KEY_PRESS_EVENT] =
     g_signal_new ("key-press-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
@@ -704,7 +819,15 @@ clutter_stage_class_init (ClutterStageClass *klass)
 		  clutter_marshal_VOID__BOXED,
 		  G_TYPE_NONE, 1,
 		  CLUTTER_TYPE_EVENT);
-  stage_signals[BUTTON_RELEASE_EVENT] =
+  /**
+   * ClutterStage::key-release-event:
+   * @stage: the actor which received the event
+   * @event: a #ClutterKeyEvent
+   *
+   * The ::key-release-event is emitted each time a keyboard button
+   * is released on @stage.
+   */
+  stage_signals[KEY_RELEASE_EVENT] =
     g_signal_new ("key-release-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
@@ -713,6 +836,14 @@ clutter_stage_class_init (ClutterStageClass *klass)
 		  clutter_marshal_VOID__BOXED,
 		  G_TYPE_NONE, 1,
 		  CLUTTER_TYPE_EVENT);
+  /**
+   * ClutterStage::motion-event:
+   * @stage: the actor which received the event
+   * @event: a #ClutterMotionEvent
+   *
+   * The ::motion-event is emitted each time the mouse pointer is
+   * moved on @stage.
+   */
   stage_signals[MOTION_EVENT] =
     g_signal_new ("motion-event",
 		  G_TYPE_FROM_CLASS (gobject_class),
@@ -730,12 +861,19 @@ static void
 clutter_stage_init (ClutterStage *self)
 {
   ClutterStagePrivate *priv;
+
+  /* a stage is a top-level object */
+  CLUTTER_SET_PRIVATE_FLAGS (self, CLUTTER_ACTOR_IS_TOPLEVEL);
   
   self->priv = priv = CLUTTER_STAGE_GET_PRIVATE (self);
 
   priv->want_offscreen  = FALSE;
   priv->want_fullscreen = FALSE;
   priv->hide_cursor     = FALSE;
+  priv->is_foreign_xwin = FALSE;
+
+  priv->xwin = None;
+  priv->gl_context = None;
 
   priv->xwin_width  = 100;
   priv->xwin_height = 100;
@@ -757,8 +895,8 @@ clutter_stage_init (ClutterStage *self)
  * calls to clutter_stage_get_default() will return the same
  * instance, with its reference count increased.
  *
- * Return value: the main #ClutterStage.  Use g_object_unref()
- *   when finished using it.
+ * Return value: the main #ClutterStage.  You should never
+ *   destroy the returned actor.
  */
 ClutterActor *
 clutter_stage_get_default (void)
@@ -768,6 +906,7 @@ clutter_stage_get_default (void)
   if (!stage_singleton)
     {
       stage_singleton = g_object_new (CLUTTER_TYPE_STAGE, NULL);
+      
       retval = CLUTTER_ACTOR (stage_singleton);
     }
   else
@@ -810,7 +949,7 @@ clutter_stage_get_xwindow (ClutterStage *stage)
  *
  * Target the #ClutterStage to use an existing external X Window.
  *
- * Return Value: TRUE if foreign window valid, FALSE otherwise 
+ * Return value: TRUE if foreign window valid, FALSE otherwise
  **/
 gboolean
 clutter_stage_set_xwindow_foreign (ClutterStage *stage,
@@ -824,6 +963,12 @@ clutter_stage_set_xwindow_foreign (ClutterStage *stage,
   Window root_return;
   Status status;
   ClutterGeometry geom;
+  ClutterStagePrivate *priv;
+
+  g_return_val_if_fail (CLUTTER_IS_STAGE (stage), FALSE);
+  g_return_val_if_fail (xid != None, FALSE);
+
+  priv = stage->priv;
 
   clutter_util_trap_x_errors();
 
@@ -837,23 +982,26 @@ clutter_stage_set_xwindow_foreign (ClutterStage *stage,
 			 &border,
 			 &depth);
 	       
-  if (clutter_util_untrap_x_errors() || !status 
-      || width == 0 || height == 0 || depth != stage->priv->xvisinfo->depth)
-    return FALSE;
+  if (clutter_util_untrap_x_errors() || !status ||
+      width == 0 || height == 0 ||
+      depth != priv->xvisinfo->depth)
+    {
+      return FALSE;
+    }
 
-  clutter_actor_unrealize (CLUTTER_ACTOR(stage));
+  clutter_actor_unrealize (CLUTTER_ACTOR (stage));
 
-  stage->priv->xwin = xid;
+  priv->xwin = xid;
+  priv->is_foreign_xwin = TRUE;
 
   geom.x = x;
   geom.y = y;
+  geom.width  = priv->xwin_width  = width;
+  geom.height = priv->xwin_height = height;
 
-  geom.width  = stage->priv->xwin_width  = width;
-  geom.height = stage->priv->xwin_height = height;
+  clutter_actor_set_geometry (CLUTTER_ACTOR (stage), &geom);
 
-  clutter_actor_set_geometry (CLUTTER_ACTOR(stage), &geom);
-
-  clutter_actor_realize (CLUTTER_ACTOR(stage));
+  clutter_actor_realize (CLUTTER_ACTOR (stage));
 
   return TRUE;
 }
@@ -1076,4 +1224,3 @@ clutter_stage_get_actor_at_pos (ClutterStage *stage,
 
   return found;
 }
-
