@@ -29,13 +29,22 @@
  *
  * #ClutterTexture is a base class for displaying and manipulating pixel
  * buffer type data.
+ *
+ * The #clutter_texture_set_from_data and #clutter_texture_set_pixbuf are
+ * used to copy image data into texture memory and subsequently realize the
+ * the texture. Unrealizing/hiding frees image data from texture memory moving
+ * to main system memory. Re-realizing then performs the opposite operation.
+ * This process allows basic management of commonly limited available texture
+ * memory.  
  */
 
 #include "clutter-texture.h"
 #include "clutter-main.h"
 #include "clutter-marshal.h"
+#include "clutter-feature.h"
 #include "clutter-util.h"
-#include "clutter-private.h" 	/* for DBG */
+#include "clutter-private.h"
+#include "clutter-debug.h"
 
 #include <GL/glx.h>
 #include <GL/gl.h>
@@ -48,26 +57,29 @@ G_DEFINE_TYPE (ClutterTexture, clutter_texture, CLUTTER_TYPE_ACTOR);
 #define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
 #endif
 
-typedef struct ClutterTextureTileDimention
+typedef struct ClutterTextureTileDimension
 {
   gint pos, size, waste;
 }
-ClutterTextureTileDimention;
+ClutterTextureTileDimension;
 
-struct ClutterTexturePrivate
+struct _ClutterTexturePrivate
 {
-  GdkPixbuf                   *pixbuf; 
   gint                         width, height;
   GLenum                       pixel_format;
   GLenum                       pixel_type;
+  GLenum                       target_type; 
+
+  GdkPixbuf                   *local_pixbuf; /* non video memory copy */
 
   gboolean                     sync_actor_size;
   gint                         max_tile_waste;
   guint                        filter_quality;
   gboolean                     repeat_x, repeat_y; /* non working */
+
   
   gboolean                     tiled;
-  ClutterTextureTileDimention *x_tiles, *y_tiles;
+  ClutterTextureTileDimension *x_tiles, *y_tiles;
   gint                         n_x_tiles, n_y_tiles;
   GLuint                      *tiles;
 };
@@ -88,25 +100,22 @@ enum
 
 enum
 {
-  SIGNAL_SIZE_CHANGE,
-  SIGNAL_PIXBUF_CHANGE,
+  SIZE_CHANGE,
+  PIXBUF_CHANGE,
   LAST_SIGNAL
 };
 
 static int texture_signals[LAST_SIGNAL] = { 0 };
 
-static void
-init_tiles (ClutterTexture *texture);
-
 static gboolean
 can_create (int    width, 
 	    int    height,
 	    GLenum pixel_format,
 	    GLenum pixel_type)
 {
-  GLint new_width;
+  GLint new_width = 0;
 
-  CLUTTER_DBG("checking %ix%i", width, height);
+  CLUTTER_NOTE (TEXTURE, "checking %ix%i", width, height);
 
   glTexImage2D (GL_PROXY_TEXTURE_2D, 0, GL_RGBA,
                 width, height, 0 /* border */,
@@ -120,11 +129,24 @@ can_create (int    width,
   return new_width != 0;
 }
 
+static gboolean
+can_create_rect_arb (int    width, 
+		     int    height,
+		     GLenum pixel_format,
+		     GLenum pixel_type)
+{
+  /* FIXME: How to correctly query what max size of NPOTS text can be */
+  if (width > 4096 || height > 4096)
+    return FALSE;
+
+  return TRUE;
+}
+
 static int
 tile_dimension (int                          to_fill,
 		int                          start_size,
 		int                          waste,
-		ClutterTextureTileDimention *tiles)
+		ClutterTextureTileDimension *tiles)
 {
   int pos     = 0;
   int n_tiles = 0;
@@ -159,7 +181,7 @@ tile_dimension (int                          to_fill,
 }
 
 static void
-init_tiles (ClutterTexture *texture)
+texture_init_tiles (ClutterTexture *texture)
 {
   ClutterTexturePrivate *priv;
   gint                   x_pot, y_pot;
@@ -173,11 +195,15 @@ init_tiles (ClutterTexture *texture)
 	   && (x_pot - priv->width < priv->max_tile_waste) 
 	   && (y_pot - priv->height < priv->max_tile_waste)))
     {
-      CLUTTER_DBG("x_pot:%i - width:%i < max_waste:%i", 
-		  x_pot, priv->width, priv->max_tile_waste);
+      CLUTTER_NOTE (TEXTURE, "x_pot:%i - width:%i < max_waste:%i",
+		    x_pot,
+		    priv->width,
+		    priv->max_tile_waste);
       
-      CLUTTER_DBG("y_pot:%i - height:%i < max_waste:%i", 
-		  y_pot, priv->height, priv->max_tile_waste);
+      CLUTTER_NOTE (TEXTURE, "y_pot:%i - height:%i < max_waste:%i",
+		    y_pot,
+		    priv->height,
+		    priv->max_tile_waste);
       
       if (x_pot > y_pot)
 	x_pot /= 2;
@@ -190,7 +216,7 @@ init_tiles (ClutterTexture *texture)
 
   priv->n_x_tiles = tile_dimension (priv->width, x_pot, 
 				    priv->max_tile_waste, NULL);
-  priv->x_tiles = g_new (ClutterTextureTileDimention, priv->n_x_tiles);
+  priv->x_tiles = g_new (ClutterTextureTileDimension, priv->n_x_tiles);
   tile_dimension (priv->width, x_pot, priv->max_tile_waste, priv->x_tiles);
 
   if (priv->y_tiles)
@@ -198,13 +224,15 @@ init_tiles (ClutterTexture *texture)
 
   priv->n_y_tiles = tile_dimension (priv->height, y_pot, 
 				    priv->max_tile_waste, NULL);
-  priv->y_tiles = g_new (ClutterTextureTileDimention, priv->n_y_tiles);
+  priv->y_tiles = g_new (ClutterTextureTileDimension, priv->n_y_tiles);
   tile_dimension (priv->height, y_pot, priv->max_tile_waste, priv->y_tiles);
 
-  CLUTTER_DBG("x_pot:%i, width:%i, y_pot:%i, height: %i max_waste:%i, "
-              " n_x_tiles: %i, n_y_tiles: %i",
-	      x_pot, priv->width, y_pot, priv->height, priv->max_tile_waste,
-	      priv->n_x_tiles, priv->n_y_tiles);
+  CLUTTER_NOTE (TEXTURE,
+                "x_pot:%i, width:%i, y_pot:%i, height: %i "
+		"max_waste:%i, n_x_tiles: %i, n_y_tiles: %i",
+		x_pot, priv->width, y_pot, priv->height,
+		priv->max_tile_waste,
+		priv->n_x_tiles, priv->n_y_tiles);
 
 }
 
@@ -240,10 +268,19 @@ texture_render_to_gl_quad (ClutterTexture *texture,
 
   if (!priv->tiled)
     {
-      glBindTexture(GL_TEXTURE_2D, priv->tiles[0]);
+      glBindTexture(priv->target_type, priv->tiles[0]);
 
-      tx = (float) priv->width / clutter_util_next_p2 (priv->width);  
-      ty = (float) priv->height / clutter_util_next_p2 (priv->height);
+      if (priv->target_type == GL_TEXTURE_2D) /* POT */
+	{
+	  tx = (float) priv->width / clutter_util_next_p2 (priv->width);  
+	  ty = (float) priv->height / clutter_util_next_p2 (priv->height);
+	}
+      else
+	{
+	  tx = (float) priv->width;
+	  ty = (float) priv->height;
+
+	}
 
       qx1 = x1; qx2 = x2;
       qy1 = y1; qy2 = y2;
@@ -266,13 +303,14 @@ texture_render_to_gl_quad (ClutterTexture *texture,
 	{
 	  int actual_w, actual_h;
 
-	  glBindTexture(GL_TEXTURE_2D, priv->tiles[i]);
+	  glBindTexture(priv->target_type, priv->tiles[i]);
 	 
 	  actual_w = priv->x_tiles[x].size - priv->x_tiles[x].waste;
 	  actual_h = priv->y_tiles[y].size - priv->y_tiles[y].waste;
 
-	  CLUTTER_DBG("rendering text tile x: %i, y: %i - %ix%i", 
-		      x, y, actual_w, actual_h);
+	  CLUTTER_NOTE (TEXTURE,
+                        "rendering text tile x: %i, y: %i - %ix%i",
+			x, y, actual_w, actual_h);
 
 	  tx = (float) actual_w / priv->x_tiles[x].size;
 	  ty = (float) actual_h / priv->y_tiles[y].size;
@@ -299,33 +337,21 @@ texture_render_to_gl_quad (ClutterTexture *texture,
 }
 
 static void
-clutter_texture_unrealize (ClutterActor *actor)
+texture_free_gl_resources (ClutterTexture *texture)
 {
-  ClutterTexture        *texture;
   ClutterTexturePrivate *priv;
 
-  texture = CLUTTER_TEXTURE(actor);
   priv = texture->priv;
 
-  if (priv->tiles == NULL)
-    return;
-
-  CLUTTER_MARK();
-
-  clutter_threads_enter ();
-
-  /* Free up texture memory */
-  if (!priv->tiled)
-    glDeleteTextures(1, priv->tiles);
-  else
-    glDeleteTextures(priv->n_x_tiles * priv->n_y_tiles, priv->tiles);
-
-  clutter_threads_leave ();
-
   CLUTTER_MARK();
 
   if (priv->tiles)
     {
+      if (!priv->tiled)
+	glDeleteTextures(1, priv->tiles);
+      else
+	glDeleteTextures(priv->n_x_tiles * priv->n_y_tiles, priv->tiles);
+
       g_free(priv->tiles);
       priv->tiles = NULL;
     }
@@ -341,12 +367,16 @@ clutter_texture_unrealize (ClutterActor *actor)
       g_free(priv->y_tiles);
       priv->y_tiles = NULL;
     }
-
-  CLUTTER_DBG("Texture unrealized");
 }
 
 static void
-clutter_texture_sync_pixbuf (ClutterTexture *texture)
+texture_upload_data (ClutterTexture *texture,
+		     const guchar   *data,
+		     gboolean        has_alpha,
+		     gint            width,
+		     gint            height,
+		     gint            rowstride,
+		     gint            bpp)
 {
   ClutterTexturePrivate *priv;
   int                    x, y, i = 0;
@@ -354,16 +384,13 @@ clutter_texture_sync_pixbuf (ClutterTexture *texture)
 
   priv = texture->priv;
 
-  g_return_if_fail (priv->pixbuf != NULL);
-
+  g_return_if_fail (data != NULL);
 
   CLUTTER_MARK();
 
   if (!priv->tiled)
     {
-      /* Single Texture 
-       *	 
-      */
+      /* Single Texture */
       if (!priv->tiles)
 	{
 	  priv->tiles = g_new (GLuint, 1);
@@ -371,61 +398,71 @@ clutter_texture_sync_pixbuf (ClutterTexture *texture)
 	  create_textures = TRUE;
 	}
 
-      CLUTTER_DBG("syncing for single tile");
+      CLUTTER_NOTE (TEXTURE, "syncing for single tile");
 
-      glBindTexture(GL_TEXTURE_2D, priv->tiles[0]);
+      glBindTexture(priv->target_type, priv->tiles[0]);
 
-      glTexParameteri(GL_TEXTURE_2D, 
+      glTexParameteri(priv->target_type, 
 		      GL_TEXTURE_WRAP_S, 
 		      priv->repeat_x ? GL_REPEAT : GL_CLAMP_TO_EDGE);
 
-      glTexParameteri(GL_TEXTURE_2D, 
+      glTexParameteri(priv->target_type, 
 		      GL_TEXTURE_WRAP_T, 
 		      priv->repeat_y ? GL_REPEAT : GL_CLAMP_TO_EDGE);
 
-      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 
+      priv->filter_quality = 1;
+
+      glTexParameteri(priv->target_type, GL_TEXTURE_MAG_FILTER, 
 		      priv->filter_quality ? GL_LINEAR : GL_NEAREST);
 
-      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
+      glTexParameteri(priv->target_type, GL_TEXTURE_MIN_FILTER, 
 		      priv->filter_quality ? GL_LINEAR : GL_NEAREST);
 
-      glPixelStorei (GL_UNPACK_ROW_LENGTH, 
-		     gdk_pixbuf_get_width(priv->pixbuf));
-      glPixelStorei (GL_UNPACK_ALIGNMENT, 
-		     gdk_pixbuf_get_n_channels (priv->pixbuf));
+      glPixelStorei (GL_UNPACK_ROW_LENGTH, priv->width);
+      glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
 	  
       if (create_textures)
 	{
-	  /* NOTE: Change to GL_RGB for non alpha textures */
-	  glTexImage2D(GL_TEXTURE_2D, 
+	  gint width, height;
+
+	  width  = priv->width;
+	  height = priv->height;
+
+	  if (priv->target_type == GL_TEXTURE_2D) /* POT */
+	    {
+	      width  = clutter_util_next_p2(priv->width);
+	      height = clutter_util_next_p2(priv->height);
+	    }
+
+	  glTexImage2D(priv->target_type, 
 		       0, 
-		       (gdk_pixbuf_get_n_channels (priv->pixbuf) == 4) ? 
-		              GL_RGBA : GL_RGB,
-		       clutter_util_next_p2(priv->width),
-		       clutter_util_next_p2(priv->height),
+		       /* (has_alpha) ? GL_RGBA : GL_RGB, */
+		       GL_RGBA,
+		       width,
+		       height,
 		       0, 
 		       priv->pixel_format, 
 		       priv->pixel_type, 
 		       NULL);
 	}
-
-      glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0,
+      
+      glTexSubImage2D (priv->target_type, 0, 0, 0,
 		       priv->width,
 		       priv->height,
 		       priv->pixel_format, 
 		       priv->pixel_type, 
-		       gdk_pixbuf_get_pixels(priv->pixbuf));
-
+		       data);
       return;
     }
 
   /* Multiple tiled texture */
   
-  CLUTTER_DBG("syncing for multiple tiles for %ix%i pixbuf",
-	      priv->width, priv->height);
+  CLUTTER_NOTE (TEXTURE,
+                "syncing for multiple tiles for %ix%i pixbuf",
+		priv->width, priv->height);
 
   g_return_if_fail (priv->x_tiles != NULL && priv->y_tiles != NULL);
-
+  
   if (priv->tiles == NULL)
     {
       priv->tiles = g_new (GLuint, priv->n_x_tiles * priv->n_y_tiles);
@@ -436,20 +473,16 @@ clutter_texture_sync_pixbuf (ClutterTexture *texture)
   for (x=0; x < priv->n_x_tiles; x++)
     for (y=0; y < priv->n_y_tiles; y++)
       {
-	GdkPixbuf *pixtmp;
-	int src_h, src_w;
+	guchar    *tmp;
+	gint        src_h, src_w, dy;
 	
 	src_w = priv->x_tiles[x].size;
 	src_h = priv->y_tiles[y].size;
 	
-	pixtmp 
-	  = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
-			   (gdk_pixbuf_get_n_channels (priv->pixbuf) == 4) ? 
-			   TRUE : FALSE,
-			   8,
-			   priv->x_tiles[x].size,
-			   priv->y_tiles[y].size);
-	
+	/* fixme - gslice ? */
+	tmp = g_malloc (sizeof(guchar) * priv->x_tiles[x].size 
+			  * priv->y_tiles[y].size * bpp);
+
 	/* clip */
 	if (priv->x_tiles[x].pos + src_w > priv->width)
 	  {
@@ -461,21 +494,20 @@ clutter_texture_sync_pixbuf (ClutterTexture *texture)
 	    src_h = priv->height - priv->y_tiles[y].pos;
 	  }
 
-	CLUTTER_DBG("copying tile %i,%i - %ix%i to 0,0 %ix%i", 
-		    priv->x_tiles[x].pos,
-		    priv->y_tiles[y].pos,
-		    src_w,
-		    src_h,
-		    priv->x_tiles[x].size,
-		    priv->y_tiles[y].size);
+	CLUTTER_NOTE (TEXTURE,
+                      "copying tile %i,%i - %ix%i to 0,0 %ix%i",
+		      priv->x_tiles[x].pos, priv->y_tiles[y].pos,
+		      src_w, src_h,
+		      priv->x_tiles[x].size,
+		      priv->y_tiles[y].size);
 
-	gdk_pixbuf_copy_area(priv->pixbuf, 
-			     priv->x_tiles[x].pos,
-			     priv->y_tiles[y].pos,
-			     src_w,
-			     src_h,
-			     pixtmp,
-			     0,0);
+	for (dy = 0; dy < src_h; dy++)
+	  {
+	    memcpy (tmp + (dy * src_w * bpp),
+		    data + ((priv->y_tiles[y].pos + dy) * rowstride)
+		         + (priv->x_tiles[x].pos * bpp),
+		    (src_w * bpp));
+	  }
 
 #ifdef CLUTTER_DUMP_TILES
 	{
@@ -489,96 +521,143 @@ clutter_texture_sync_pixbuf (ClutterTexture *texture)
 	}
 #endif
 
-	glBindTexture(GL_TEXTURE_2D, priv->tiles[i]);
+	glBindTexture(priv->target_type, priv->tiles[i]);
 
-	glTexParameteri(GL_TEXTURE_2D, 
+	glTexParameteri(priv->target_type, 
 			GL_TEXTURE_WRAP_S, 
 			priv->repeat_x ? GL_REPEAT : GL_CLAMP_TO_EDGE);
 
-	glTexParameteri(GL_TEXTURE_2D, 
+	glTexParameteri(priv->target_type, 
 			GL_TEXTURE_WRAP_T, 
 			priv->repeat_y ? GL_REPEAT : GL_CLAMP_TO_EDGE);
 
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 
+	glTexParameteri(priv->target_type, GL_TEXTURE_MAG_FILTER, 
 			priv->filter_quality ? GL_LINEAR : GL_NEAREST);
 
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
+	glTexParameteri(priv->target_type, GL_TEXTURE_MIN_FILTER, 
 			priv->filter_quality ? GL_LINEAR : GL_NEAREST);
 
 	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-	
-	glPixelStorei (GL_UNPACK_ROW_LENGTH, gdk_pixbuf_get_width(pixtmp));
-	glPixelStorei (GL_UNPACK_ALIGNMENT, 
-		       gdk_pixbuf_get_n_channels (priv->pixbuf));
+
+	glPixelStorei (GL_UNPACK_ROW_LENGTH, src_w);
+	glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
 	
 	if (create_textures)
 	  {
-
-	    glTexImage2D(GL_TEXTURE_2D, 
+	    glTexImage2D(priv->target_type, 
 			 0, 
-			 (gdk_pixbuf_get_n_channels (priv->pixbuf) == 4) ? 
+			 (bpp == 4) ? 
 			 GL_RGBA : GL_RGB,
 			 priv->x_tiles[x].size,
 			 priv->y_tiles[y].size,
 			 0, 
 			 priv->pixel_format, 
 			 priv->pixel_type, 
-			 gdk_pixbuf_get_pixels(pixtmp));
+			 tmp);
 	  }
 	else 
 	  {
 	    /* Textures already created, so just update whats inside 
 	    */
-	    glTexSubImage2D (GL_TEXTURE_2D, 0, 
+	    glTexSubImage2D (priv->target_type, 0, 
 			     0, 0,
 			     priv->x_tiles[x].size,
 			     priv->y_tiles[y].size,
 			     priv->pixel_format, 
 			     priv->pixel_type, 
-			     gdk_pixbuf_get_pixels(pixtmp));
+			     tmp);
 	  }
 
-	g_object_unref(pixtmp);
+	g_free(tmp);
 
 	i++;
       }
 }
 
 static void
-clutter_texture_realize (ClutterActor *actor)
+clutter_texture_unrealize (ClutterActor *actor)
 {
-  ClutterTexture *texture;
+  ClutterTexture        *texture;
+  ClutterTexturePrivate *priv;
 
   texture = CLUTTER_TEXTURE(actor);
+  priv = texture->priv;
+
+  if (priv->tiles == NULL)
+    return;
 
   CLUTTER_MARK();
 
-  if (texture->priv->pixbuf == NULL)
+  /* Move image data from video to main memory 
+  */
+  if (priv->local_pixbuf == NULL)
+    priv->local_pixbuf = clutter_texture_get_pixbuf (texture);
+  
+  texture_free_gl_resources (texture);
+
+  CLUTTER_NOTE (TEXTURE, "Texture unrealized");
+}
+
+static void
+clutter_texture_realize (ClutterActor *actor)
+{
+  ClutterTexture       *texture;
+  ClutterTexturePrivate *priv;
+
+  texture = CLUTTER_TEXTURE(actor);
+  priv = texture->priv;
+
+  CLUTTER_MARK();
+
+  if (priv->local_pixbuf != NULL)
     {
-      /* Dont allow realization with no pixbuf */
-      CLUTTER_DBG("*** Texture has no pixbuf cannot realize ***");
-      CLUTTER_DBG("*** flags %i ***", actor->flags);
+      /* Move any local image data we have from unrealization  
+       * back into video memory.	 
+      */
+      clutter_texture_set_pixbuf (texture, priv->local_pixbuf);
+      g_object_unref (priv->local_pixbuf);
+      priv->local_pixbuf = NULL;
+    }
+  else
+    {
+      /* Dont allow realization with no pixbuf - note set_pixbuf/data 
+       * will set realize flags.	 
+      */
+      CLUTTER_NOTE (TEXTURE,
+                    "Texture has no image data cannot realize");
+
+      CLUTTER_NOTE (TEXTURE, "flags %i", actor->flags);
       CLUTTER_ACTOR_UNSET_FLAGS (actor, CLUTTER_ACTOR_REALIZED);
-      CLUTTER_DBG("*** flags %i ***", actor->flags);
+      CLUTTER_NOTE (TEXTURE, "flags %i", actor->flags);
       return;
     }
-  CLUTTER_DBG("Texture realized");
 
-  if (texture->priv->tiled)
-    init_tiles(texture);
-
-  clutter_texture_sync_pixbuf (texture);
+  CLUTTER_NOTE (TEXTURE, "Texture realized");
 }
 
 static void
 clutter_texture_show (ClutterActor *self)
 {
+  ClutterActorClass *parent_class;
+
+  /* chain up parent show */
+  parent_class = CLUTTER_ACTOR_CLASS (clutter_texture_parent_class);
+  if (parent_class->show)
+    parent_class->show (self);
+
   clutter_actor_realize (self);
 }
 
 static void
 clutter_texture_hide (ClutterActor *self)
 {
+  ClutterActorClass *parent_class;
+
+  /* chain up parent hide */
+  parent_class = CLUTTER_ACTOR_CLASS (clutter_texture_parent_class);
+  if (parent_class->hide)
+    parent_class->hide (self);
+
   clutter_actor_unrealize (self);
 }
 
@@ -589,24 +668,26 @@ clutter_texture_paint (ClutterActor *self)
   gint            x1, y1, x2, y2;
   guint8          opacity;
 
-  CLUTTER_DBG("@@@ for '%s' @@@", 
-	      clutter_actor_get_name(self) ? 
-                          clutter_actor_get_name(self) : "unknown");
+  CLUTTER_NOTE (PAINT,
+                "@@@ for '%s' @@@",
+		clutter_actor_get_name (self) ? clutter_actor_get_name (self)
+                                                         : "unknown");
   glPushMatrix();
 
   glEnable(GL_BLEND);
-  glEnable(GL_TEXTURE_2D);
-  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+  glEnable(texture->priv->target_type);
+  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
 
   opacity = clutter_actor_get_opacity(self);
 
-  CLUTTER_DBG("setting opacity to %i\n", opacity);
+  CLUTTER_NOTE (PAINT, "setting opacity to %i\n", opacity);
   glColor4ub(255, 255, 255, opacity);
 
   clutter_actor_get_coords (self, &x1, &y1, &x2, &y2);
-  texture_render_to_gl_quad (texture, x1, y1, x2, y2);
+  /* Paint will of translated us */
+  texture_render_to_gl_quad (texture, 0, 0, x2 - x1, y2 - y1);
 
-  glDisable(GL_TEXTURE_2D);
+  glDisable(texture->priv->target_type);
   glDisable(GL_BLEND);
 
   glPopMatrix();
@@ -615,19 +696,19 @@ clutter_texture_paint (ClutterActor *self)
 static void 
 clutter_texture_dispose (GObject *object)
 {
-  ClutterTexture *self = CLUTTER_TEXTURE(object);
+  ClutterTexture *texture = CLUTTER_TEXTURE(object);
   ClutterTexturePrivate *priv;
 
-  priv = self->priv;
+  priv = texture->priv;
 
   if (priv != NULL)
     {
-      clutter_actor_unrealize (CLUTTER_ACTOR(self));
+      texture_free_gl_resources (texture);
 
-      if (priv->pixbuf != NULL)
+      if (priv->local_pixbuf != NULL)
 	{
-	  g_object_unref (priv->pixbuf);
-	  priv->pixbuf = NULL;
+	  g_object_unref (priv->local_pixbuf);
+	  priv->local_pixbuf = NULL;
 	}
     }
 
@@ -668,7 +749,13 @@ clutter_texture_set_property (GObject      *object,
       break;
     case PROP_USE_TILES:
       priv->tiled = g_value_get_boolean (value);
-      CLUTTER_DBG("Texture is tiled ? %i", priv->tiled);
+
+      if (priv->target_type == GL_TEXTURE_RECTANGLE_ARB
+	  && priv->tiled == TRUE)
+	priv->target_type  = GL_TEXTURE_2D;
+
+      CLUTTER_NOTE (TEXTURE, "Texture is tiled ? %s",
+		    priv->tiled ? "yes" : "no");
       break;
     case PROP_MAX_TILE_WASTE:
       priv->max_tile_waste = g_value_get_int (value);
@@ -712,7 +799,11 @@ clutter_texture_get_property (GObject    *object,
   switch (prop_id) 
     {
     case PROP_PIXBUF:
-      g_value_set_pointer (value, priv->pixbuf);
+      {
+	GdkPixbuf *pixb;
+	pixb = clutter_texture_get_pixbuf (texture);
+	g_value_set_pointer (value, pixb);
+      }
       break;
     case PROP_USE_TILES:
       g_value_set_boolean (value, priv->tiled);
@@ -770,7 +861,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
      g_param_spec_pointer ("pixbuf",
 			   "Pixbuf source for Texture.",
 			   "Pixbuf source for Texture.",
-			   G_PARAM_READWRITE));
+			   CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_USE_TILES,
@@ -778,8 +869,13 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 			   "Enable use of tiled textures",
 			   "Enables the use of tiled GL textures to more "
 			   "efficiently use available texture memory",
-			   TRUE,
-			   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+			   /* FIXME: This default set at runtime :/  
+                            * As tiling depends on what GL features available.
+                            * Need to figure out better solution
+			   */
+			   (clutter_feature_available 
+			     (CLUTTER_FEATURE_TEXTURE_RECTANGLE) == FALSE),
+			   G_PARAM_CONSTRUCT_ONLY | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_SYNC_SIZE,
@@ -788,7 +884,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 			   "Auto sync size of actor to underlying pixbuf"
 			   "dimentions",
 			   TRUE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_REPEAT_X,
@@ -797,7 +893,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 			   "Reapeat underlying pixbuf rather than scale" 
 			   "in x direction. Currently UNWORKING",
 			   FALSE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_REPEAT_Y,
@@ -806,7 +902,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 			   "Reapeat underlying pixbuf rather than scale" 
 			   "in y direction. Currently UNWORKING",
 			   FALSE,
-			   G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+			   G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   /* FIXME: Ideally this option needs to have some kind of global
    *        overide as to imporve performance.
@@ -822,7 +918,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 		       0,
 		       G_MAXINT,
 		       1,
-		       G_PARAM_CONSTRUCT | G_PARAM_READWRITE));
+		       G_PARAM_CONSTRUCT | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_MAX_TILE_WASTE,
@@ -834,7 +930,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 		       0,
 		       G_MAXINT,
 		       64,
-		       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+		       G_PARAM_CONSTRUCT_ONLY | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_PIXEL_TYPE,
@@ -844,7 +940,7 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 		       0,
 		       G_MAXINT,
 		       PIXEL_TYPE,
-		       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+		       G_PARAM_CONSTRUCT_ONLY | CLUTTER_PARAM_READWRITE));
 
   g_object_class_install_property
     (gobject_class, PROP_PIXEL_FORMAT,
@@ -854,9 +950,19 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 		       0,
 		       G_MAXINT,
 		       GL_RGBA,
-		       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+		       G_PARAM_CONSTRUCT_ONLY | CLUTTER_PARAM_READWRITE));
 
-  texture_signals[SIGNAL_SIZE_CHANGE] =
+  /**
+   * ClutterTexture::size-change:
+   * @texture: the texture which received the signal
+   * @width: the width of the new texture
+   * @height: the height of the new texture
+   *
+   * The ::size-change signal is emitted each time the size of the
+   * pixbuf used by @texture changes.  The new size is given as
+   * argument to the callback.
+   */
+  texture_signals[SIZE_CHANGE] =
     g_signal_new ("size-change",
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
@@ -865,8 +971,14 @@ clutter_texture_class_init (ClutterTextureClass *klass)
 		  clutter_marshal_VOID__INT_INT,
 		  G_TYPE_NONE, 
 		  2, G_TYPE_INT, G_TYPE_INT);
-
-  texture_signals[SIGNAL_PIXBUF_CHANGE] =
+  /**
+   * ClutterTexture::pixbuf-change:
+   * @texture: the texture which received the signal
+   * 
+   * The ::pixbuf-change signal is emitted each time the pixbuf
+   * used by @texture changes.
+   */
+  texture_signals[PIXBUF_CHANGE] =
     g_signal_new ("pixbuf-change",
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
@@ -882,14 +994,8 @@ clutter_texture_init (ClutterTexture *self)
 {
   ClutterTexturePrivate *priv;
 
-  priv        = g_new0 (ClutterTexturePrivate, 1);
+  priv = g_new0 (ClutterTexturePrivate, 1);
 
-  /* FIXME: It seems defaults via props do not get set
-   * on all props for sub classes ( ie labels ).
-   * Should they be ?
-   *
-   * Setting them here also to be sure + safe. 
-  */
   priv->max_tile_waste = 64;
   priv->filter_quality = 0;
   priv->tiled        = TRUE;
@@ -897,24 +1003,238 @@ clutter_texture_init (ClutterTexture *self)
   priv->pixel_format = GL_RGBA;
   priv->repeat_x     = FALSE;
   priv->repeat_y     = FALSE;
-  priv->pixbuf = NULL;
 
+  if (clutter_feature_available (CLUTTER_FEATURE_TEXTURE_RECTANGLE))
+    {
+      priv->target_type  = GL_TEXTURE_RECTANGLE_ARB;
+      priv->tiled        = FALSE;
+    }
+  else
+    priv->target_type = GL_TEXTURE_2D;
 
   self->priv  = priv;
 }
 
+static void
+pixbuf_destroy_notify (guchar  *pixels, gpointer data)
+{
+  g_free (pixels);
+}
+
 /**
  * clutter_texture_get_pixbuf:
  * @texture: A #ClutterTexture
  *
- * Gets the underlying #GdkPixbuf for the #ClutterTexture.
+ * Gets a #GdkPixbuf representation of the #ClutterTexture data. 
+ * The created #GdkPixbuf is not owned by the texture but the caller.  
  *
- * Return value: The underlying #GdkPixbuf
+ * Return value: A #GdkPixbuf
  **/
 GdkPixbuf*
 clutter_texture_get_pixbuf (ClutterTexture* texture)
 {
-  return texture->priv->pixbuf;
+  ClutterTexturePrivate *priv;
+  GdkPixbuf             *pixbuf = NULL;
+  guchar                *pixels = NULL;
+
+  priv = texture->priv;
+
+  if (priv->tiles == NULL)
+    return NULL; 
+
+  if (!priv->tiled)
+    {
+      pixels = g_malloc (priv->width * priv->height * 4);
+      
+      if (pixels == NULL)
+	return NULL;
+
+      glBindTexture(priv->target_type, priv->tiles[0]);
+
+      glPixelStorei (GL_UNPACK_ROW_LENGTH, priv->width);
+      glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
+
+      /* read data from gl text and return as pixbuf */
+      glGetTexImage (priv->target_type,
+		     0,
+		     priv->pixel_format,
+		     priv->pixel_type,
+		     (GLvoid*)pixels);
+
+      pixbuf = gdk_pixbuf_new_from_data ((const guchar*)pixels,
+					 GDK_COLORSPACE_RGB,
+					 (priv->pixel_format == GL_RGBA),
+					 8,
+					 priv->width,
+					 priv->height,
+					 priv->width * 4,
+					 pixbuf_destroy_notify,
+					 NULL);
+    }
+  else
+    {
+      int x,y,i;
+
+      i = 0;
+
+      pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
+			       (priv->pixel_format == GL_RGBA),
+			       8,
+			       priv->width,
+			       priv->height);
+
+      for (x=0; x < priv->n_x_tiles; x++)
+	for (y=0; y < priv->n_y_tiles; y++)
+	  {
+	    GdkPixbuf  *tmp_pixb;
+	    gint        src_h, src_w;
+	
+	    src_w = priv->x_tiles[x].size;
+	    src_h = priv->y_tiles[y].size;
+
+	    pixels = g_malloc (src_w * src_h *4);
+
+	    glBindTexture(priv->target_type, priv->tiles[i]);
+	    
+	    glPixelStorei (GL_UNPACK_ROW_LENGTH, src_w);
+	    glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
+
+	    glGetTexImage (priv->target_type,
+			   0,
+			   priv->pixel_format,
+			   priv->pixel_type,
+			   (GLvoid*)pixels);
+	
+	    tmp_pixb 
+	      = gdk_pixbuf_new_from_data ((const guchar*)pixels,
+					  GDK_COLORSPACE_RGB,
+					  (priv->pixel_format == GL_RGBA),
+					  8,
+					  src_w,
+					  src_h,
+					  src_w * 4,
+					  pixbuf_destroy_notify,
+					  NULL);
+	
+	    gdk_pixbuf_copy_area (tmp_pixb,
+				  0,
+				  0,
+				  src_w,
+				  src_h,
+				  pixbuf,
+				  priv->x_tiles[x].pos,
+				  priv->x_tiles[y].pos);
+
+	    g_object_unref (tmp_pixb);
+
+	    i++;
+	  }
+      
+    }
+
+  return pixbuf;
+}
+
+/**
+ * clutter_texture_set_from_data:
+ * @texture: A #ClutterTexture
+ * @data: Image data in RGB type colorspace.
+ * @has_alpha: Set to TRUE if image data has a alpha channel.
+ * @width: Width in pixels of image data.
+ * @height: Height in pixels of image data
+ * @rowstride: Distance in bytes between row starts.
+ * @bpp: bytes per pixel ( Currently only 4 supported )
+ *
+ * Sets #ClutterTexture image data.
+ *
+ * Since 0.2. This function is likely to change in future versions.
+ **/
+void
+clutter_texture_set_from_data (ClutterTexture *texture,
+			       const guchar   *data,
+			       gboolean        has_alpha,
+			       gint            width,
+			       gint            height,
+			       gint            rowstride,
+			       gint            bpp)
+{
+  ClutterTexturePrivate *priv;
+  gboolean               texture_dirty = TRUE;
+
+  priv = texture->priv;
+
+  g_return_if_fail (data != NULL);
+  g_return_if_fail (bpp == 4);
+
+  /* FIXME: check other image props */
+  texture_dirty = (width != priv->width || height != priv->height);
+
+  priv->width  = width;
+  priv->height = height;
+
+  if (has_alpha)
+    priv->pixel_format = GL_RGBA;
+  else
+    priv->pixel_format = GL_RGB;
+
+  if (texture_dirty)
+    {
+      texture_free_gl_resources (texture);
+
+      if (priv->tiled == FALSE)
+	{
+	  if (priv->target_type == GL_TEXTURE_RECTANGLE_ARB
+	      && !can_create_rect_arb (priv->width, 
+				       priv->height,
+				       priv->pixel_format,
+				       priv->pixel_type))
+	    {
+	      /* If we cant create NPOT tex of this size fall back to tiles */
+	      priv->tiled = TRUE; 
+	      priv->target_type = GL_TEXTURE_2D;
+	    }
+	  else if (priv->target_type == GL_TEXTURE_2D
+		   && !can_create(clutter_util_next_p2(priv->width), 
+				  clutter_util_next_p2(priv->height),
+				  priv->pixel_format, 
+				  priv->pixel_type))
+	    { 
+	      priv->tiled = TRUE; 
+	    }
+	}
+
+      /* Figure our tiling etc */
+      if (priv->tiled)
+	texture_init_tiles (texture); 
+    }
+
+  CLUTTER_NOTE (TEXTURE, "set size %ix%i\n",
+		priv->width,
+		priv->height);
+
+  texture_upload_data (texture, data, has_alpha, 
+		       width, height, rowstride, bpp);
+
+  CLUTTER_ACTOR_SET_FLAGS (CLUTTER_ACTOR(texture), CLUTTER_ACTOR_REALIZED);
+
+  if (texture_dirty)
+    {
+      g_signal_emit (texture, texture_signals[SIZE_CHANGE], 
+		     0, priv->width, priv->height);
+
+      if (priv->sync_actor_size)
+	clutter_actor_set_size (CLUTTER_ACTOR(texture), 
+				priv->width, 
+				priv->height);
+    }
+  
+  /* rename signal */
+  g_signal_emit (texture, texture_signals[PIXBUF_CHANGE], 0); 
+
+  /* If resized actor may need resizing but paint() will do this */
+  if (CLUTTER_ACTOR_IS_MAPPED (CLUTTER_ACTOR(texture)))
+    clutter_actor_queue_redraw (CLUTTER_ACTOR(texture));
+
 }
 
 /**
@@ -922,98 +1242,25 @@ clutter_texture_get_pixbuf (ClutterTexture* texture)
  * @texture: A #ClutterTexture
  * @pixbuf: A #GdkPixbuf
  *
- * Sets the underlying #GdkPixbuf for the #ClutterTexture
+ * Sets a  #ClutterTexture image data from a #GdkPixbuf
  *
  **/
 void
 clutter_texture_set_pixbuf (ClutterTexture *texture, GdkPixbuf *pixbuf)
 {
   ClutterTexturePrivate *priv;
-  gboolean               texture_dirty = TRUE;
 
   priv = texture->priv;
 
   g_return_if_fail (pixbuf != NULL);
 
-  if (priv->pixbuf != NULL)
-    {
-      texture_dirty = (gdk_pixbuf_get_width (pixbuf) 
-		       != gdk_pixbuf_get_width (priv->pixbuf)
-		       ||
-		       gdk_pixbuf_get_height (pixbuf) 
-		       != gdk_pixbuf_get_height (priv->pixbuf)
-		       ||
-		       gdk_pixbuf_get_n_channels (pixbuf)
-		       != gdk_pixbuf_get_n_channels (priv->pixbuf));
-
-      g_object_unref(priv->pixbuf);
-
-      /* If the actual pixbuf has changed size/format destroy
-       * existing textures ready for recreation. If
-       * size matches we can reuse. 
-      */
-      if (texture_dirty)
-	{
-	  clutter_actor_unrealize (CLUTTER_ACTOR(texture));
-	}
-      else
-	{
-	  /* If texture realised sync things for change */
-	  if (CLUTTER_ACTOR_IS_REALIZED(CLUTTER_ACTOR(texture)))
-	    {
-	      priv->pixbuf = pixbuf; 
-
-	      /* FIXME: better locking stratergy 
-	      */
-	      clutter_threads_enter();
-	      clutter_texture_sync_pixbuf (texture);
-	      clutter_threads_leave();
-	    }
-	}
-    }
-
-  clutter_threads_enter();
-      
-  priv->pixbuf = pixbuf; 
-  priv->width  = gdk_pixbuf_get_width (pixbuf);
-  priv->height = gdk_pixbuf_get_height (pixbuf);
-
-  g_object_ref (pixbuf);
-
-  if (gdk_pixbuf_get_n_channels (pixbuf) == 3)
-    priv->pixel_format = GL_RGB;
-  else
-    priv->pixel_format = GL_RGBA;
-
-  /* Force tiling if pixbuf is too big for single texture */
-  if (priv->tiled == FALSE 
-      && texture_dirty 
-      && !can_create(clutter_util_next_p2(priv->width), 
-		     clutter_util_next_p2(priv->height),
-		     priv->pixel_format, 
-		     priv->pixel_type))
-    priv->tiled = TRUE;
-  clutter_threads_leave();
-
-  if (priv->sync_actor_size)
-    clutter_actor_set_size (CLUTTER_ACTOR(texture), 
-			      priv->width, 
-			      priv->height);
-
-  CLUTTER_DBG("set size %ix%i\n", priv->width, priv->height);
-
-  if (texture_dirty)
-    g_signal_emit (texture, texture_signals[SIGNAL_SIZE_CHANGE], 
-		   0, priv->width, priv->height);
-
-  if (priv->tiled && texture_dirty)
-      init_tiles (texture); 
-
-  g_signal_emit (texture, texture_signals[SIGNAL_PIXBUF_CHANGE], 0); 
-
-  /* If resized actor may need resizing but paint() will do this */
-  if (CLUTTER_ACTOR_IS_MAPPED (CLUTTER_ACTOR(texture)))
-    clutter_actor_queue_redraw (CLUTTER_ACTOR(texture));
+  clutter_texture_set_from_data (texture,
+				 gdk_pixbuf_get_pixels (pixbuf),
+				 gdk_pixbuf_get_has_alpha (pixbuf),
+				 gdk_pixbuf_get_width (pixbuf),
+				 gdk_pixbuf_get_height (pixbuf),
+				 gdk_pixbuf_get_rowstride (pixbuf),
+				 4);
 }
 
 /**
@@ -1060,7 +1307,7 @@ clutter_texture_new (void)
  * Gets the size in pixels of the untransformed underlying texture pixbuf data.
  *
  **/
-void
+void 				/* FIXME: rename to get_image_size */
 clutter_texture_get_base_size (ClutterTexture *texture, 
 			       gint           *width,
 			       gint           *height)
@@ -1096,7 +1343,7 @@ clutter_texture_bind_tile (ClutterTexture *texture, gint index)
   ClutterTexturePrivate *priv;
 
   priv = texture->priv;
-  glBindTexture(GL_TEXTURE_2D, priv->tiles[index]);
+  glBindTexture(priv->target_type, priv->tiles[index]);
 }
 
 /**
